// project
#include "IsoSurface.h"
#include "VecData.h"
#include "render_utils.h"
#define BRANCH_TOL 0.2

/** calc isosurface
 *
 * \param m      model
 * \param dat    data displayed on model
 * \param v      isovalue
 * \param member true if part of model
 * \param t      time
 * \param branch branch cut information (NULL if none)
 */
void
IsoSurface::calc(Model* m, DATA_TYPE *dat, double v, 
		       vector<bool>&member, int t, double *branch )
{
  double brmin = branch ? branch[0]+BRANCH_TOL*(branch[1]-branch[0]) : 0;
  double brmax = branch ? branch[1]-BRANCH_TOL*(branch[1]-branch[0]) : 0;
  EdgePtMap ep_map;
  PPoint *pt = _p;
  _tm        = t;
  _val       = v;

#pragma omp parallel for shared(ep_map,pt)
  for( int i=0; i<m->numVol(); i++ ) {
    if( member[i] ) {

      // ignore elements crossing the branch cut
      if( branch ) {
        const int* nodes= m->_vol[i]->obj();
        bool low=false, high=false;
        for( int n=0; n<m->_vol[i]->ptsPerObj(); n++ ) {
          if( dat[nodes[n]] <= brmin ) low=true;
          if( dat[nodes[n]] >= brmax ) high=true;
        }
        if( low && high ) 
          continue;
      }

      int npoly;
      MultiPoint **lpoly = m->_vol[i]->isosurf( dat, _val, npoly, pt, ep_map );

      if (lpoly != NULL){
        for( int j=0; j<npoly; j++ )
        {
#pragma omp critical
          _ele.push_back(static_cast<SurfaceElement*>(lpoly[j]));
        }
        delete[] lpoly;
      }
    }
  }
  determine_vert_norms(*_p);
}


void 
IsoSurface::draw(bool recolor, bool flat) {
  glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );
  zsort( NULL, 1, false );
  GLfloat *vb0 = _rd.buffer();
  if( recolor ) {
    GLfloat *buff = _rd.buffer();
    for( int i=0; i<_rd.numVtx(); i++ )
      memcpy( buff+VBO_COLOR_OFFSET+VBO_ELEM_SZ*i, _fillcolor, 4*sizeof(GLfloat) );
  } else if( fresh() ){
    GLfloat* vbo = _rd.buffer();
    buffer( _fillcolor, NULL, NULL, NULL, vbo, flat );
  }
  if( recolor || fresh() )
    _rd.update_nodalbuff();
  render();
}

IsoSurface::~IsoSurface()
{
  for( auto& ele : _ele ) {
    delete ele->pt();
    delete ele;
  }
  //delete _vertnorm;
  delete _p;
}


/** output the surface as an auxilliary grid
 *
 * \param basename base filename
 *
 * \pre the files have been opened and the number of elements and points 
 *      written to it
*/
void 
IsoSurface::saveAux( string basename, int pt_offset ) 
{
  string ptfile = basename+".pts_t";
  FILE*pout = fopen( ptfile.c_str(), "a" );
  const GLfloat *p = _p->pt();
  for( int i=0; i<_p->num(); i++, p+=3 )
    fprintf( pout, "%.1f %.1f %.1f\n", p[0], p[1], p[2] );
  fclose( pout );

  string elemfile = basename+".elem_t";
  FILE *eout = fopen( elemfile.c_str(), "a" );
  for( auto const& poly: _ele ) {
    const int* n = poly->obj();
    for( int j=0; j<poly->ptsPerObj(); j++ ){
      fprintf( pout, "%d ", n[j]+pt_offset );
    }
    fprintf( pout, "\n" );
  }
  fclose( eout );
}

void 
IsoSurface::clear( GLfloat *n )
{
  if ( !_ele.empty() ) 
    for ( int i=0; i<_ele.size(); i++ ) {
      delete _ele[i];
    }
  _ele.clear();
  _zlist.clear();
  _vertnorm.reset();
  delete _p;
  _p = new PPoint;
  _rd.clear(); 
  _fresh = true;
}


