#include "Surfaces.h"
#include <string.h>
#include "VecData.h"
#include <algorithm>
#include <vector>
#include <iterator>
#include <set>
#include "legacyGL.h"
/** find the z depth of a point (assume w=1)
 *
 * \param m row of the ModelViewProjection 
 * \param v vertex (x,y,z)
 *
 * \return the z depth buffer coordinate
 */
inline GLfloat 
z_proj( const GLfloat* m, const GLfloat *v )
{
  GLfloat z=m[3];
  for( int i=0; i<3; i++ ) z += m[i]*v[i];
  return z;
}


/**
 * @brief  set the material colour properties
 */
void
Surfaces :: set_material( )
{
  GLfloat diffuseb[4], specularb[4];

  CSET( diffuseb,  _diffuse[0]*_backlight,  _diffuse[3]  );
  CSET( specularb, _specular[0]*_backlight, _specular[3] );
}


void Surfaces :: fillcolor( float r, float g, float b, float a )
{
  _fillcolor[0] = r;
  _fillcolor[1] = g;
  _fillcolor[2] = b;
  _fillcolor[3] = a;
}


//! set outline colour
void Surfaces :: outlinecolor( float r, float g, float b, float a )
{
  _outlinecolor[0] = r;
  _outlinecolor[1] = g;
  _outlinecolor[2] = b;
  _outlinecolor[3] = a;
}


/** set the vertex normals for the surface
 *
 * \param pt all points
 *
 */
void Surfaces::determine_vert_norms( PPoint& pt )
{
  vector<bool> has_norm(pt.num());  // if elements attached to node

  GLfloat *tvn = (GLfloat*)calloc( pt.num()*3, sizeof(GLfloat) );
  memset(  tvn, 0, pt.num()*3*sizeof(GLfloat) );

  const GLfloat* n;
  for ( auto ele : _ele ) {
    if ( (n=ele->nrml(0)) == NULL ) continue;
    const int *pnt = ele->obj();
    for ( int j=0; j<ele->ptsPerObj(); j++ ) {
      for ( int k=0; k<3; k++ ) tvn[3*pnt[j]+k] += n[k];
      has_norm[pnt[j]] = true;
    }
  }
  // count \# nodes in surface
  for ( int i=0; i<pt.num(); i++ ) {
    if ( has_norm[i] ) {
      normalize(tvn+i*3);
      _vert.insert(i);
    }
  }
  _vertnorm.reset(tvn);

  //find sharp edges
  const float max_normdiff = 0.8;
  for ( auto ele : _ele ) {
    if ( (n=ele->nrml(0)) == NULL ) continue;
    const int    *pnt = ele->obj();
    vector<bool> sharp(ele->ptsPerObj());
    const int    np   = ele->ptsPerObj(); 
    for ( int j=0; j<np; j++ ) {
      if( dot( n, _vertnorm.get()+pnt[j]*3 ) < max_normdiff )
        sharp[j] = true;
    }
    for ( int j=0; j<np; j++ ) {
      if( sharp[j] && sharp[(j+1)%np] ) {
        if( pnt[j] < pnt[(j+1)%np] ) 
          _sharp_edges.insert( EdgePair( pnt[j], pnt[(j+1)%np] ) );
        else
          _sharp_edges.insert( EdgePair( pnt[(j+1)%np], pnt[j] ) );
      }
    }
  }
}


/**
 * @brief sort elements by z depth
 *
 * @param stride element drawing stride
 * @param sort   do we really sort?
 */
void Surfaces::zsort( void *context, int stride, bool sort )
{
  _zlist.resize( (_ele.size()+stride-1)/stride );

  if( sort ) {
    // build modelview projection matrix - only compute row to determine z
    GLfloat mvp[4]{};
    GLfloat *proj=ProjectionMat[context].m, *modview=ModelViewMat[context].m;
    for( int i=0; i<4; i++ )
      for(int j=0; j<4; j++ )
        mvp[i] += modview[2+j*4]*proj[4*i+j]; // column major order

    if( memcmp(mvp, _oldmvp, sizeof(mvp) ) || stride != _oldstride) {
      memcpy( _oldmvp, mvp, sizeof(mvp) );
    }
#pragma omp parallel for 
    for ( int i=0; i<_ele.size(); i+=stride ){
      const PPoint *pts = _ele[i]->pt();
      _zlist[i] = {i,0,_index};
      const int *nn     = _ele[i]->obj();
      for( int k=0; k<3; k++ )
        _zlist[i].z += z_proj(mvp,pts->pt(nn[k]));
    }
    std::sort( _zlist.begin(), _zlist.end(),[](const vtx_z a, const vtx_z b){return a.z<b.z;} );

  } else {
    for ( int i=0; i<_ele.size(); i+=stride )
      _zlist[i/stride] = { i, 0, _index };
  }
  _oldstride = stride;
}


/**
 * @brief determine element data by averaging the vertex data
 *
 * @param ptdata data on vertices
 *
 * @return the data for the elements
 */
vector<GLfloat> 
Surfaces::to_elem(GLfloat *ptdata)
{
  vector<GLfloat> edata(_ele.size());
#pragma omp parallel for
  for( int e=0; e<_ele.size(); e++ ) {
    const int *n = _ele[e]->obj();
    for( int j=0; j<_ele[e]->ptsPerObj(); j++ )
      edata[e] += ptdata[n[j]];
    edata[e] /= (float)_ele[e]->ptsPerObj();
  }
  return edata;
}


/** buffer the surface
 *  \param fill     fill colour
 *  \param cs       colour scale
 *  \param dat      if flat, element data, else nodal data (NULL for nodata display)
 *  \param dataopac data opacity
 *  \param buffer   buffer in which to place data
 *  \param cmode    colouring mode
 *
 *  \pre  the elements need to have been sorted by calling ::sort()
 *  \post the buffer reference points just past the entry added
 */
void Surfaces::buffer( GLfloat *fill, Colourscale *cs, DATA_TYPE *dat,
                     dataOpac* dataopac, GLfloat *&buffer, char cmode )
{
  for ( auto &a :_zlist )
    buffer_elem( a.i, fill, cs, dat, dataopac, buffer, cmode );
  _zlist.clear();
}


/** buffer opaque surface elements
 *
 *  \param opaque   transparency limit for opaque
 *  \param fill     fill colour
 *  \param cs       colour scale
 *  \param dat      nodal data (NULL for nodata display)
 *  \param dataopac data opacity
 *  \param buffer   buffer in which to place data
 *  \param cmode    colouring mode
 *
 *  \post the buffer reference points just past the last entry added
 *  \post opaque elements are removed from \p _zlist
 */
void Surfaces::buffer_opaque( float opaque, GLfloat *fill, Colourscale *cs, DATA_TYPE *dat,
                     dataOpac* dataopac, GLfloat *&buffer, char cmode )
{
  set<int> solid;

  for ( int j=0; j<_zlist.size(); j++ ) {
    vtx_z          &a = _zlist[j];
    SurfaceElement *s = _ele[a.i];
    const int      *n = s->obj();
    bool      eleopac = true;
    for( int i=0; i<s->ptsPerObj(); i++ )
      if( dataopac->alpha(dat[n[i]]) < opaque ) {
        eleopac = false;
        break;
      }
    if( eleopac ) {
      solid.insert(j);
      buffer_elem( a.i, fill, cs, dat, dataopac, buffer, cmode );
    }
  }
  for( auto i=solid.rbegin(); i != solid.rend(); i++ )
    _zlist.erase(_zlist.begin()+(*i));
}


/** draw an element in the surface
 * \param  ele      the element index
 *  \param fill     fill colour
 *  \param cs       colour scale
 *  \param dat      nodal or elemental data (NULL for nodata display)
 *  \param dataopac data opacity
 *  \param buf[out] buffer to fill
 *
 *  \pre the elements need to have been sorted by calling ::sort()
 *  \post \p buf is advanced by the number of floats written
 */
void Surfaces::buffer_elem( int ele, GLfloat *fill, Colourscale *cs, DATA_TYPE *dat,
                            dataOpac* dataopac, GLfloat *&buf, char cmode )
{
  const GLfloat *nrml = (cmode&CM_FLAT)?_ele[ele]->nrml():_vertnorm.get();
  GLfloat *dp = dat;
  if( dat && cmode&CM_ELEM )
    dp += _sele[ele];
  else if( dat && cmode&CM_FLAT )
    dp += ele;
  buf += _ele[ele]->buffer( 0, fill, cs, dp, dataopac, nrml, buf, _index, cmode )*VBO_ELEM_SZ*3;
}


/** draw the sharp edges of the object
 *
 * \param buff VBO buffer
 */
void
Surfaces::buffer_sharp_edges( GLfloat *&buff )
{
  GLfloat nrml[3] = { 0., 0., 0. };
  GLfloat col[4]  = { 1, 1, 1, 0.1 };
  for( auto edge : _sharp_edges )  {
    buff += buff_vtx( buff, _p->pt(edge.first),  nrml, col, 0 );
    buff += buff_vtx( buff, _p->pt(edge.second), nrml, col, 0 );
  }
}


/** redraw elements through which the branch cut passes with flat shading 
 *
 * \param range  range of cut
 * \param data   data to display
 * \param cs     colour scale
 * \param stride output stride
 * \param opac   data opacity    
 */
void 
Surfaces ::correct_branch_elements( GLdouble *range, DATA_TYPE *data,
                         Colourscale *cs, int stride, dataOpac *opac )
{
#define BRANCH_TOL  0.2
  bool cross_branch( DATA_TYPE *d, int n, double min, double max, double tol );

  DATA_TYPE d[MAX_NUM_SURF_NODES*2];
  for ( int i=0; i<_ele.size(); i+=stride ) {
    const int *n = _ele[i]->obj();
    for( int j=0; j<_ele[i]->ptsPerObj(); j++ )
      d[j] = data[n[j]];
    if( cross_branch( d, _ele[i]->ptsPerObj(), range[0], range[1], BRANCH_TOL ) )
      assert(0);
  }
}


/** flip the normals
 */
void Surfaces :: flip_norms()
{
  for( size_t i=0; i<_p->num()*3; i++ )
    _vertnorm[i] *= -1.;
}


/** write the surface to a file 
 * \param of output stream
 */
void
Surfaces :: to_file( ofstream &of )
{
    of << num() << "  " << label() << endl;
 
    for( auto &e : _ele ) {
      const int* n=e->obj();
      if( e->ptsPerObj()==3 ) 
        of << "Tr";
      else
        of << "Qd";
      for( int k=0; k< e->ptsPerObj(); k++ )
        of << " " << n[k];
      of << endl;
    }  
}


/** count the number of equivalent triangles 
 *
 *  \post _numtri is updated
 */
void
Surfaces :: count_tris()
{
  _numtri=0;
  for( auto &e: _ele )
    _numtri += e->ntris();
}
