#ifdef __APPLE__
#  define GL_SILENCE_DEPRECATION
#  include <OpenGL/gl3.h> 
#else
#  if defined(WIN32)
#    define GLEW_STATIC 1
#  endif
# include <GL/glew.h>
#endif

#include "render_func.h"
#include <string>
#include <iostream>
#include <sstream>
#include <set>
#include "VecData.h"

const float POINT_BACK_LIMIT = -0.95;

#define ICOSHX .525731112119133606 
#define ICOSHZ .850650808352039932
static GLfloat icoshVtx[12][3] = {    
   {-ICOSHX, 0.0, ICOSHZ}, {ICOSHX, 0.0, ICOSHZ}, {-ICOSHX, 0.0, -ICOSHZ}, {ICOSHX, 0.0, -ICOSHZ},    
   {0.0, ICOSHZ, ICOSHX}, {0.0, ICOSHZ, -ICOSHX}, {0.0, -ICOSHZ, ICOSHX}, {0.0, -ICOSHZ, -ICOSHX},    
   {ICOSHZ, ICOSHX, 0.0}, {-ICOSHZ, ICOSHX, 0.0}, {ICOSHZ, -ICOSHX, 0.0}, {-ICOSHZ, -ICOSHX, 0.0} 
};
static unsigned int icoshTri[20][3] = { 
   {0,4,1}, {0,9,4}, {9,5,4}, {4,5,8}, {4,8,1},    
   {8,10,1}, {8,3,10}, {5,3,8}, {5,2,3}, {2,7,3},    
   {7,10,3}, {7,6,10}, {7,11,6}, {11,0,6}, {0,1,6}, 
   {6,1,10}, {9,0,11}, {9,11,2}, {9,2,5}, {7,2,11} };



int generate_shader_program(const char* vertex_source, const char* fragment_source)
{
  // build and compile our shader program
  // ------------------------------------
  // vertex shader
  int vertexShader = glCreateShader(GL_VERTEX_SHADER);
  glShaderSource(vertexShader, 1, &vertex_source, NULL);
  glCompileShader(vertexShader);
  // check for shader compile errors
  int success;
  char infoLog[512];

  glGetShaderiv(vertexShader, GL_COMPILE_STATUS, &success);
  if (!success) {
    glGetShaderInfoLog(vertexShader, 512, NULL, infoLog);
    std::cout << "ERROR::SHADER::VERTEX::COMPILATION_FAILED\n" << infoLog << std::endl;
  }

  // fragment shader
  int fragmentShader = glCreateShader(GL_FRAGMENT_SHADER);
  glShaderSource(fragmentShader, 1, &fragment_source, NULL);
  glCompileShader(fragmentShader);
  // check for shader compile errors
  glGetShaderiv(fragmentShader, GL_COMPILE_STATUS, &success);
  if (!success)
  {
    glGetShaderInfoLog(fragmentShader, 512, NULL, infoLog);
    std::cout << "ERROR::SHADER::FRAGMENT::COMPILATION_FAILED\n" << infoLog << std::endl;
  }
  // link shaders
  int shaderProgram = glCreateProgram();
  glAttachShader(shaderProgram, vertexShader);
  glAttachShader(shaderProgram, fragmentShader);
  glLinkProgram(shaderProgram);
  // check for linking errors
  glGetProgramiv(shaderProgram, GL_LINK_STATUS, &success);
  if (!success) {
    glGetProgramInfoLog(shaderProgram, 512, NULL, infoLog);
    std::cout << "ERROR::SHADER::PROGRAM::LINKING_FAILED\n" << infoLog << std::endl;
  }

  glDeleteShader(vertexShader);
  glDeleteShader(fragmentShader);

  return shaderProgram;
}


/**
 * @brief copy vertex data to GPU
 *
 * @param unbind unbind buffer after transfer?
 */
void 
render_data::init_nodalbuff(bool unbind)
{
  glBindBuffer(GL_ARRAY_BUFFER, gl_nodalbuff);
  if( nodalbuff.size() != _old_sz ) {
    glBufferData(GL_ARRAY_BUFFER, nodalbuff.size()*sizeof(float), nodalbuff.data(), GL_DYNAMIC_DRAW);
    _old_sz = nodalbuff.size();
  } else 
    glBufferSubData(GL_ARRAY_BUFFER, 0, nodalbuff.size()*sizeof(float), nodalbuff.data());

  if(unbind)
    glBindBuffer(GL_ARRAY_BUFFER, 0);
}


/**
 * @brief do all the ugly setting up for OPenGL including
 * defining the vertex buffer and binding to OGL context
 *
 * \param c  openGL context
 * \param pt draw simulated 3D points
 *
 * \return true iff the initialization was operformed
 */
bool 
render_data::setup_render_data(void *c, bool pt)
{
  if(!init) {
#if !defined(__APPLE__) && !defined(OSMESA) 
    glewExperimental = GL_TRUE;
    GLenum err = glewInit();
    if (GLEW_OK != err) {
      std::cerr << "GLEW problem -- aborting!" << std::endl;
    }
#endif
    if( c ) _context=c;
    shader_default = get_default_lighting_shader();

    shader_wireframe = pt ? get_lit_point_shader() : get_wireframe_shader();

    set_identity_4x4_mat(tfm);

    glGenVertexArrays(1, &gl_vertexarray);
    glBindVertexArray(gl_vertexarray);
    glGenBuffers(1, &gl_nodalbuff);

    init_nodalbuff(false);

    glEnableVertexAttribArray(0);
    glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, sizeof(float) * VBO_ELEM_SZ, (void*)(VBO_XYZ_OFFSET));
    glEnableVertexAttribArray(1);
    glVertexAttribPointer(1, 3, GL_FLOAT, GL_FALSE, sizeof(float) * VBO_ELEM_SZ, (void*)(VBO_NORM_OFFSET*sizeof(float)));
    glEnableVertexAttribArray(2);
    glVertexAttribPointer(2, 3, GL_FLOAT, GL_FALSE, sizeof(float) * VBO_ELEM_SZ, (void*)(VBO_COLOR_OFFSET*sizeof(float)));
    glEnableVertexAttribArray(3);
    glVertexAttribPointer(3, 1, GL_FLOAT, GL_FALSE, sizeof(float) * VBO_ELEM_SZ, (void*)((VBO_COLOR_OFFSET+3)*sizeof(float)));
    glEnableVertexAttribArray(4);
    glVertexAttribPointer(4, 1, GL_FLOAT, GL_FALSE, sizeof(float) * VBO_ELEM_SZ, (void*)(VBO_ENT_OFFSET*sizeof(float)));

    glBindVertexArray(0);
    glBindBuffer(GL_ARRAY_BUFFER, 0);
    init = true;
    return true;
  } else
    return false;
}



render_data::~render_data()
{
  if( init ) {
    glDeleteBuffers( 1,  &gl_nodalbuff );
    glDeleteVertexArrays( 1, &gl_vertexarray );
    glDeleteProgram(shader_default);
    glDeleteProgram(shader_wireframe);
  }
}


void
render_data::set_clipping(int shader_program) 
{
  glUseProgram(shader_program);

  // now feed in the uniform input data (mostly the transformation matrices)
  // into the shader program
  set_shader_view_matrices(shader_program,_context);
  set_shader_mat4(shader_program, "model", &tfm[0][0]);

  // set up lighting
  GLfloat light_colour[4]= { 1.0f, 1.0f, 1.0f, 1.0f };
  V3f ld = light_dir(_context).Rotate( V3f(0,0,1) );
  set_shader_vec3(shader_program,  "lightDir",         ld.e          );
  set_shader_vec3(shader_program,  "lightColor",       light_colour  );
  set_shader_float(shader_program, "ambientStrength",  ambient       );
  set_shader_float(shader_program, "diffuseStrength",  diffuse       );
  set_shader_float(shader_program, "specularStrength", specular      );
  set_shader_int(shader_program,   "backLit",          back_lit      );
 
  GLfloat noclip[4] = { 0.f, 0.f, 0.f, 1000.f };
  for( int i=0; i<NUM_CP; i++ ) {
    GLfloat *clipEqn = clipping(i)&&!_on_top ? CLIP_EQN[i]: noclip;
    std::ostringstream clipEqnStr;
    clipEqnStr << "clipEqn[" << i << "]";
    set_shader_vec4(shader_program, clipEqnStr.str().c_str(), clipEqn);
  }
  
  // set up material properties for each surface
  for( int i=0; i<MAX_NUM_SURFS; i++ ){
    if( material[i][0] == BOGUS_VAL ) continue;
    ostringstream materialStr;
    materialStr << "materialProps[" << i << "]";
    set_shader_vec4(shader_program, materialStr.str().c_str(), material[i]);
  }
}


/** finish rendering */
void
render_data :: finish_render()
{
  // unbind render data and shader
  glBindVertexArray(0);
  glBindBuffer(GL_ARRAY_BUFFER, 0 );
  glUseProgram(0);

  // Forces the execution of OpenGL commands
  //glFlush();
  _dirty = false;
}


/** render the surface elements
 *
 * \param wireframe   render wireframe else solid elements
 *
 * \pre \p ntris is the number of triangles to render
 * \pre \p n_opaque is the number of opaque vertices (3/tri)
 * \pre \p nodalbuff is filled first with opaque vertices, followed by translucent
 * \pre \p material[i] has the info for surface[i]
 */
void 
RenderTris:: mesh_render( bool wireframe )
{
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

  glBindVertexArray(gl_vertexarray);
  glBindBuffer(GL_ARRAY_BUFFER, gl_nodalbuff);
  int shader_program = wireframe ? shader_wireframe : shader_default;
  if( wireframe ) {
    glPolygonMode( GL_FRONT_AND_BACK, GL_LINE );
    shader_program = shader_wireframe;
    glLineWidth(lineWidth);
  } else {
    shader_program = shader_default;
    glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );
  }
  set_clipping(shader_program);

  //run the rendering pipeline, first opaque then translucent
  glDisable(GL_BLEND);
  glDepthMask(GL_TRUE);
  if( _on_top ) 
    glClear(GL_DEPTH_BUFFER_BIT);
  
  glDrawArrays(GL_TRIANGLES, 0, n_opaque);

  GLint ntrans_vtx = _n*3 - n_opaque;
  if(ntrans_vtx) {
    glDepthMask(GL_FALSE);
    glEnable(GL_BLEND);
    glDrawArrays(GL_TRIANGLES, n_opaque, ntrans_vtx );
  }

  finish_render();
}
/** render the surface elements
 *
 * \param wireframe   render wireframe else solid elements
 *
 * \pre \p ntris is the number of triangles to render
 * \pre \p n_opaque is the number of opaque vertices (3/tri)
 * \pre \p nodalbuff is filled first with opaque vertices, followed by translucent
 * \pre \p material[i] has the info for surface[i]
 */
void 
RenderLines:: mesh_render( bool wf )
{
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

  glBindVertexArray(gl_vertexarray);
  glBindBuffer(GL_ARRAY_BUFFER, gl_nodalbuff);
  int shader_program;
  if( threeD() ) {
    shader_program = shader_default;
    glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );
  } else {
    glPolygonMode( GL_FRONT_AND_BACK, GL_LINE );
    shader_program = shader_wireframe;
    glLineWidth(lineWidth);
  }
  set_clipping(shader_program);

  //run the rendering pipeline, first opaque then translucent
  glDisable(GL_BLEND);
  glDepthMask(GL_TRUE);
  if( !threeD() )
    glDrawArrays(GL_LINES, 0, n_opaque);
  else
    glDrawArrays(GL_TRIANGLES, 0, n_opaque);

  GLint ntrans_vtx = n_opaque==-1 ? 0 : (_n-n_opaque)*3;
  if(ntrans_vtx) {
    glDepthMask(GL_FALSE);
    glEnable(GL_BLEND);
    if( !threeD() )
      glDrawArrays(GL_LINES, n_opaque, ntrans_vtx );
    else
      glDrawArrays(GL_TRIANGLES, n_opaque, ntrans_vtx);
  }
  finish_render();
}


/**
 * @brief buffer a line
 *
 * @param lbuff buffer entries for all \p n+1 vertices 
 * @param n     number of segments in line
 *
 * \post buffer is increased in size to hold new line
 */
void
RenderLines :: buffer_line( GLfloat *lbuff, int n )
{
  int oldn = _n*VBO_ELEM_SZ;
  _n += threeD() ? _num_sides*6*(n+1) : n*2;  //#vtx += cylinders + 2 end caps
  nodalbuff.resize(_n*VBO_ELEM_SZ);
  GLfloat *vbo = nodalbuff.data() + oldn; 
  if( !threeD() ) {              
    // 1D
    for( int i=0; i<n; i++ ) {
      GLfloat *vtx0 = lbuff+i*VBO_ELEM_SZ;
      memcpy( vbo, vtx0, 2*sizeof(GLfloat)*VBO_ELEM_SZ );
      vbo += 2*VBO_ELEM_SZ;
    }  
  } else {
    for( int i=0; i<n; i++ ) {
      GLfloat *vtx0 = lbuff+i*VBO_ELEM_SZ;
      GLfloat *col0 = vtx0 + VBO_COLOR_OFFSET;
      GLfloat *next = i<n-1 ? vtx0+2*VBO_ELEM_SZ : NULL;
      buff_cyl( vbo, vtx0, vtx0+VBO_ELEM_SZ, _num_sides, lineWidth, col0, col0+VBO_ELEM_SZ, 0, i, next );
    }
  }
}


/**
 * @brief    change number of sides to draw segments
 *
 * @param n     number of sides per cylinder
 * \param extra extranumber of vertices to allocate per line segment
 *
 * @note the lines must be rebuffered for this to have an effect
 */
void
RenderLines :: sides( int n, int extra ) 
{
  if( _num_sides==n ) return;
  _num_sides = n<3 ? 0 : n;
  _dirty = true;
}


/**
 * @brief add two buffers together
 *
 * @param rl buffer to append
 *
 * \note if not specified, assume all lines are opaque
 */
RenderLines& 
RenderLines :: operator += (RenderLines &rl)
{
  if( threeD() != rl.threeD() ) throw 1;
  if( n_opaque == -1 )    n_opaque    = _n;
  if( rl.n_opaque == -1 ) rl.n_opaque = rl._n;
  nodalbuff.insert( _n==n_opaque?nodalbuff.end():nodalbuff.begin()+n_opaque*VBO_ELEM_SZ,
                     rl.nodalbuff.begin(), rl.nodalbuff.begin()+rl.n_opaque*VBO_ELEM_SZ );
  nodalbuff.insert( nodalbuff.end(), rl.nodalbuff.begin()+rl.n_opaque*VBO_ELEM_SZ, rl.nodalbuff.end() );
  _n += rl._n;
  n_opaque += rl.n_opaque;
  
  return *this;
}


/**
 * @brief buffer a point
 *
 * @param vbo start of buffer
 * @param b   index of sphere in buffer
 * @param vtx vertex position
 * @param col colour
 * @param reg region
 *
 * \pre  buffer is allocated
 * \post buffer is filled, either 0 or 3D, for point \p b
 */
void
RenderSpheres :: buffer_sphere( GLfloat *vbo, int b, const GLfloat *vtx, const GLfloat *col, int reg )
{
  vbo += _esz*b; 
  if( !threeD() ) {              
    GLfloat n[3] = {0,0,1};
    vbo += buff_vtx( vbo, vtx, n, col, reg );
  } else {
    buff_sphere( vbo, vtx, _vtx.data(), _tri.size()/3, _tri.data(), _radius, col, reg );   
  }
}

/**
 * @brief define spherical approx by #subdisions of an icoshahedron
 *
 * @param nsd number of subdivisions, 0=0D, 1=20 faces, 2=80, 3=320, etc
 *
 * \post \p _tri and \p _vtx are populated for unit sphere at origin
 */
void
RenderSpheres::subdiv( int nsd )
{
  if( _nsubdiv == nsd ) return;
  if( threeD() ) _dirty = true;

  _nsubdiv=nsd;
  if( !_nsubdiv ) {
    _vtx.clear();
    _tri.clear();
    _esz = VBO_ELEM_SZ;
    prim_sz = 1;
    return;
  }
  prim_sz = mk_icosahedron( _nsubdiv, _tri, _vtx );
  _esz = prim_sz*VBO_ELEM_SZ;
}


void 
RenderSpheres:: mesh_render( bool wf )
{
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

  glBindVertexArray(gl_vertexarray);
  glBindBuffer(GL_ARRAY_BUFFER, gl_nodalbuff);
  int shader_program;
  if( threeD() ) {
    shader_program = shader_default;
    glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );
  } else {
    glPolygonMode( GL_FRONT_AND_BACK, GL_LINE );
    shader_program = shader_wireframe;
    glPointSize(_radius);
  }
  set_clipping(shader_program);
 
  //run the rendering pipeline, first opaque then translucent
  glDisable(GL_BLEND);
  glDepthMask(GL_TRUE);
  if( threeD() )
    glDrawArrays(GL_TRIANGLES, 0, n_opaque*prim_sz);
  else
    glDrawArrays(GL_POINTS, 0, n_opaque);

  int ntrans_vtx = (_n - n_opaque) * prim_sz;
  if(ntrans_vtx) {
    glDepthMask(GL_FALSE);
    glEnable(GL_BLEND);
    if( threeD() )
      glDrawArrays(GL_TRIANGLES, n_opaque*prim_sz, ntrans_vtx);
    else
      glDrawArrays(GL_POINTS, n_opaque, ntrans_vtx );
  }
  finish_render();
}


void
RenderHilights::pt( unsigned int p, const GLfloat *vtx, GLfloat rad, bool threeD )
{
  assign( _pt, vtx );
  _hiPt  = p;
  _ptRad = rad;
  if( threeD ) {
    vector<unsigned int> _sphTri;
    vector<GLfloat>      _sphVtx;
    mk_icosahedron( SPHERE_REFINE, _sphTri, _sphVtx );
    size_t osz = nodalbuff.size();
    nodalbuff.resize( osz + _sphTri.size()*VBO_ELEM_SZ );
    GLfloat *buff = nodalbuff.data() + osz;
    for( int s=0;s<_sphTri.size();s++) {
      GLfloat p[3];
      int node = _sphTri[s];
      add( scale( assign( p, _sphVtx.data()+node*3 ), _ptRad ), _pt, p );
      buff += buff_vtx( buff, p, _sphVtx.data()+node*3, _colPt, 0 );
      _nPtVtx++;
    }
  } else {
    nodalbuff.resize(VBO_ELEM_SZ) ;
    buff_vtx( nodalbuff.data(), _pt, (const GLfloat[3]){1,0,0}, _colPt, 0 );
    _nPtVtx = 1;
  }
}


void
RenderHilights::cnnx( unsigned int c, const GLfloat *vtx0, const GLfloat*vtx1, GLfloat rad, int ns )
{
  GLfloat n[3] = {1., 0, 0};
  _hiCnnx    = c;
  _cnnxRad   = rad;
  bool _cnnx3D = ns>=3;
  size_t osz = nodalbuff.size();
  if( _cnnx3D ) {
    nodalbuff.resize( osz+VBO_ELEM_SZ*ns*4*3 ); 
    GLfloat *buff = nodalbuff.data()+osz;
    _nCnnxVtx = buff_cyl( buff, vtx0, vtx1, ns, rad, _colCnnx, _colCnnx, 0, false, NULL )/VBO_ELEM_SZ;
  } else {
    nodalbuff.resize( osz+2*VBO_ELEM_SZ); 
    buff_vtx( nodalbuff.data()+osz,             vtx0,  n, _colCnnx, 0 );
    buff_vtx( nodalbuff.data()+osz+VBO_ELEM_SZ, vtx1,  n, _colCnnx, 0 );
    _nCnnxVtx=2;
  }
}


/**
 * @brief add this surface tris to render
 *
 * @param ntri     number of tris
 * @param ebuff    VBO format buffered eleemnts
 * @param fill     fill in surfele?
 * @param s        if >=0, highlight surface number
 */
void
RenderHilights::surfele( int ntri, const GLfloat* ebuff, bool fill, int s )
{
  if( s>=0 ) _hiSurf = s;
  buffer_hiEles( ntri, ebuff, fill,  s<0?_colAssSurf:_colHiSurf,
                          fill?_colSurfEdge:(s<0?_colAssSurf:_colHiSurf) );
}


/**
 * @brief Buffer a bunch of tris
 *
 * @param ntri       \#triangles
 * @param ebuff      buffer fo vertices
 * @param fill       fill the triangles?
 * @param solidCol   fill colour
 * @param outlineCol edge colour
 */
void
RenderHilights::buffer_hiEles( int ntri, const GLfloat* ebuff, bool fill, 
                             const GLfloat *solidCol, const GLfloat *outlineCol )
{
  _wireTri.resize( _wireTri.size() + ntri*3*VBO_ELEM_SZ );
  GLfloat* wbuff = _wireTri.data() + _wireTri.size() - ntri*3*VBO_ELEM_SZ;
  GLfloat *nbuff;
  if( fill ) {
    nodalbuff.resize( nodalbuff.size() + ntri*3*VBO_ELEM_SZ );
    nbuff = nodalbuff.data() + nodalbuff.size() - ntri*3*VBO_ELEM_SZ;
  }
  for( int i=0; i<ntri; i++ ){
    GLfloat a[3], b[3], n[3];
    sub( ebuff+VBO_ELEM_SZ*2, ebuff, a );
    sub( ebuff+VBO_ELEM_SZ,   ebuff, b );
    normalize( cross( a, b, n ) );
    for( int j=0; j<3; j++ ){
      wbuff += buff_vtx( wbuff, ebuff+(i*3+j)*VBO_ELEM_SZ, n, outlineCol, 0 );
      if( fill )
        nbuff += buff_vtx( nbuff, ebuff+(i*3+j)*VBO_ELEM_SZ, n, solidCol, 0 );
    }
  }
}


/**
 * @brief buffer surfaces of a volume element
 *
 * @param ntri number of triangles
 * @param vtx  vertices
 * @param fill fill in triangles?
 * @param v    if >=0, highlighted volume index
 *
 * @return 
 */
void
RenderHilights:: volele( int ntri, const GLfloat *vtx, bool fill, int v )
{
  if( v>=0 ) _hiVol = v;

  GLfloat *buff  = new GLfloat[ntri*3*VBO_ELEM_SZ];
  for( int i=0; i<ntri*3; i++ )
    memcpy( buff+i*VBO_ELEM_SZ, vtx+i*3, 3*sizeof(GLfloat) );

  buffer_hiEles( ntri, buff, fill, v<0?_colAssVol:_colHiVol, fill?_colVolEdge:(v<0?_colAssVol:_colHiVol) );

  delete[] buff;
}


/**
 * @brief render the buffer contents which are arranged as follows:
 *
 * [vtx of highlighted point if _nPtVtx==1 ]
 * [vtx0 of highlighted cnnx 
 *  vtx0 of highlighted cnnx if _nCnnxVtx==2 ]
 * [ vtx X n_filled ] filled triangles
 * [ vtx X (nodalbuff.size()/VBO_ELEM_SZ-n_filled-nPtVtx-nCnnxVtx) ] outlined triangles
 *
 * @param wf
 */
void 
RenderHilights:: mesh_render( bool wf )
{
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

  glBindVertexArray(gl_vertexarray);
  glBindBuffer(GL_ARRAY_BUFFER, gl_nodalbuff);

  glDisable(GL_BLEND);
  glDepthMask(GL_TRUE);
  int sp = 0;

  size_t offset = 0;
  if( _nPtVtx==1 ) {
    set_clipping(sp=shader_wireframe);
    glPointSize(_ptRad);
    glDrawArrays(GL_POINTS, 0, 1);
  } else  {
    set_clipping(sp=shader_default);
    glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );
    glDrawArrays(GL_TRIANGLES, offset, _nPtVtx );
  }
  offset += _nPtVtx;

  if( _nCnnxVtx==2 ) {
    if( sp != shader_wireframe ) {
      set_clipping(sp=shader_wireframe);
    }
    glLineWidth( _cnnxRad );
    glDrawArrays(GL_LINES, offset, 2);
  } else if( _nCnnxVtx>2 ) {
    if( sp != shader_default ) {
      set_clipping(sp=shader_default);
    }
    glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );
    glDrawArrays(GL_TRIANGLES, offset, _nCnnxVtx );
  }
  offset += _nCnnxVtx;
  
  if( !_line3D ) {
    if( sp != shader_wireframe ) {
      set_clipping(sp=shader_wireframe);
    }
    glLineWidth( _lineRad );
    glDrawArrays(GL_LINES, offset, _nLineVtx);
    offset += _nLineVtx;
  }

  if( sp != shader_default ) {
    set_clipping(sp=shader_default);
  }
  glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );
  glDrawArrays(GL_TRIANGLES, offset, _n_filled );
  offset += _n_filled;

  set_clipping(shader_wireframe);
  glPolygonMode( GL_FRONT_AND_BACK, GL_LINE );
  glDrawArrays(GL_TRIANGLES, offset, _wireTri.size()/VBO_ELEM_SZ );

  finish_render();
}


void
RenderHilights::clear()
{
  render_data::clear();
  _wireTri.clear();
  _n_filled  = 0;
  _ptRad     = 0.;
  _cnnxRad   = 0.;
  _nCnnxVtx  = 0;
  _nPtVtx    = 0;
}


void
RenderHilights::update_nodalbuff()
{
  _n_filled = nodalbuff.size()/VBO_ELEM_SZ;
  _n_filled -= _nPtVtx + _nCnnxVtx;
  if( !_line3D ) _n_filled -= _nLineVtx;
  nodalbuff.insert(nodalbuff.end(), _wireTri.begin(), _wireTri.end());
  render_data::update_nodalbuff();
}


void
RenderHilights:: line( unsigned int hl, int p0, int p1, const int*cnnx, const GLfloat *pt, GLfloat rad, int ns )
{
  GLfloat n[3] = {1., 0, 0};
  _hiLine  = hl;
  _lineRad = rad;
  _line3D  = ns>=3;
  size_t osz = nodalbuff.size();
  int    nseg = p1 - p0 + 1;
  GLfloat *buff;
  if( _line3D ) {
    nodalbuff.resize( osz+VBO_ELEM_SZ*(nseg*ns*2+2*ns)*3 ); 
    buff = nodalbuff.data()+osz;
    for( int p=p0; p<=p1; p++ )
      buff_cyl( buff, pt+3*cnnx[2*p], pt+3*cnnx[2*p+1], ns, rad, _colLine, _colLine, 0, p!=p0, p==p1?NULL:pt+3*cnnx[2*p+3] );
  } else {
    nodalbuff.resize( osz+2*VBO_ELEM_SZ*nseg ); 
    buff = nodalbuff.data()+osz;
    for( int p=p0; p<=p1; p++ ){
      buff += buff_vtx( buff, pt+3*cnnx[2*p],    n, _colLine, 0 );
      buff += buff_vtx( buff, pt+3*cnnx[2*p+1],  n, _colLine, 0 );
    }
  }
  _nLineVtx=(buff-nodalbuff.data()-osz)/VBO_ELEM_SZ;
}



/**
 * @brief subdivide icosahedron by a factor of 4
 *
 * \param snsubdiv[in]  \#times to subdivide,0=point,1=20,2=80,3=320,etc
 * \param tri[out]      triangles
 * \param vtx[out]      vertices
 *
 * \return number of vertices in triangles
 * \post \p tri abd \p vtx are allocated and populated
 */
int 
mk_icosahedron( int nsubdiv, vector<unsigned int>& tri, vector<GLfloat>& vtx )
{
  vtx.resize(12*3);
  tri.resize(20*3);
  for( int i=0; i<12; i++ ) 
    assign( vtx.data()+i*3, icoshVtx[i] );
  for( int i=0; i<20; i++ ) 
    assign( tri.data()+i*3, icoshTri[i] );

 for( int j=1; j<nsubdiv; j++ ) { // 4X more triangles per loop
   int ntri = tri.size()/3;
   tri.resize(ntri*4*3);
   int nvtx = vtx.size()/3;
   vtx.resize((nvtx+3*ntri)*3);
   int nloop=ntri;
   for( int i=0; i<nloop; i++ ) { 
     int a = tri[i*3];
     int b = tri[i*3+1];
     int c = tri[i*3+2];
     int d = nvtx;
     int e = d+1;
     int f = e+1;
     GLfloat *ax = vtx.data()+a*3;
     GLfloat *bx = vtx.data()+b*3;
     GLfloat *cx = vtx.data()+c*3;
     GLfloat *dx = vtx.data()+d*3;
     GLfloat *ex = dx+3;
     GLfloat *fx = ex+3;
     normalize( add( ax, bx, dx ) );
     normalize( add( bx, cx, ex ) );
     normalize( add( cx, ax, fx ) );
     tri[i*3]   = a;
     tri[i*3+1] = d;
     tri[i*3+2] = f;
     // new tris
     tri[ntri*3]   = d;
     tri[ntri*3+1] = b;
     tri[ntri*3+2] = e;
     ntri++;
     tri[ntri*3]   = e;
     tri[ntri*3+1] = c;
     tri[ntri*3+2] = f;
     ntri++;
     tri[ntri*3]   = f;
     tri[ntri*3+1] = d;
     tri[ntri*3+2] = e;
     ntri++;
     nvtx += 3;
   }
 }
 return tri.size();
}


/** fill vertex frame buffer array
 *
 * \param buff buffer to fill
 * \param xyz  point coordinates
 * \param nrml point normal
 * \param col  point colour
 * 
 * \return number of floats copied
 */
int
buff_vtx( GLfloat *buff, const GLfloat *xyz, const GLfloat *nrml, const GLfloat *col, GLint surf )
{
  memcpy( buff+VBO_XYZ_OFFSET,   xyz,    3*sizeof(GLfloat) );
  memcpy( buff+VBO_NORM_OFFSET,  nrml,   3*sizeof(GLfloat) );
  memcpy( buff+VBO_COLOR_OFFSET, col,    4*sizeof(GLfloat) );
  GLfloat surffl=surf;
  memcpy( buff+VBO_ENT_OFFSET,  &surffl, sizeof(GLfloat) );
  return VBO_ELEM_SZ;
}


/**
 * @brief  buffer a sphere
 *
 * @param buff   buffer to fill
 * @param ctr    center of sphere
 * @param vtx    point positions on unity sphere at origin
 * @param ntri   number of triangles
 * @param tris   connectivity of vertices
 * @param radius radius of sphere
 * @param col    colour of sphere
 * @param reg    region
 *
 * @return number of floats buffered
 */
int
buff_sphere( GLfloat* &buff, const GLfloat *ctr, const GLfloat* vtx, int ntri,
             const unsigned int* tris, float radius, const GLfloat *col, int reg )
{
  for( int i=0; i<ntri*3; i++ ){
    GLfloat P[3], n[3];
    scale( assign(P, assign(n, vtx+tris[i]*3)), radius );
    add( P, ctr, P );
    buff += buff_vtx( buff, P, scale(n,-1.), col, reg );
  }
  return ntri*3*VBO_ELEM_SZ;
}


/**
 * @brief render a line segment as a cylinder. 
 * If joining to previous segment, the points on the end of the previous cylinder are used,
 * otherwise points on the plane normal to the cylinder axis passing through the starting vertex are generated.
 * If there is a following segment, the intersection plane between the two segments is the average of the 2
 * end cap planes passing through the second vertex, otherwise the end cap plane is used.
 * The points from the starting end are translated along the cylinder axis to the end plane.
 *
 * @param buff   buffer to fill
 * @param vtx0   starting vertex
 * @param vtx1   ending vertex
 * @param ns     number of sides
 * @param radius radius
 * @param c0     color vertex 0 (r,g,b,a)
 * @param c1     color vertex 1 (r,g,b,a)
 * @param join   join segment with previous one, else cap
 * @param next   next vertex in line, or NULL if none
 *
 * @return number of GLfloats added to buffer
 *
 * \pre  \p buff is large enough to hold cylinder: ( 3*\p ns*( (\p join?0:1) + 2 + (\p next==NULL?1:0) )*VBO_ELEM_SZ
 * \pre  if \p join, preceding cylinder is just before \p buff
 * \post \p buff points to one past the last entry
 */
int 
buff_cyl( GLfloat* &buff, const GLfloat *vtx0, const GLfloat *vtx1, int ns, GLfloat radius, const GLfloat *c0, 
                                                   const  GLfloat *c1, int reg, bool join, const GLfloat *next )
{
  // compute cylinder axis
  GLfloat axis[3];
  normalize(sub( vtx1, vtx0, axis ));
 
  // calculate cylinder points
  GLfloat* pt[ns*2];
  GLfloat* n[ns*2];
  GLfloat  npts[ns*2][3];
  GLfloat  newn[ns*2][3];

  // do first end which may join to previous segment
  if( join ) {
    // if join, copy initial points from last segment
    GLfloat* pt0[ns];
    GLfloat* base =  buff - VBO_ELEM_SZ*(ns*6-5);
    for( int i=0; i<ns; i++ ) {
      pt[i] = base + i*6*VBO_ELEM_SZ + VBO_XYZ_OFFSET;
      // figure out the normal of the old point to the new axis
      GLfloat d[3];
      GLfloat pa[3];
      sub( pt[i], vtx0, d );
      assign( pa, axis );
      normalize( sub( d, scale( pa, dot( axis, d ) ), newn[i] ) );
      n[i]  = scale(newn[i],-1.);
    }
  } else { // make new points
    // compute orthogonal radial directions
    GLfloat r0[3]={0.,0.,0.}, 
            r1[3]={0.,0.,0.};
    if( equal(axis, (const GLfloat[]){0,0,1}, 1.e-4f )   ||
        equal(axis, (const GLfloat[]){0,0.,-1.}, 1.e-4f )  )
      r0[0] = 1.;
    else 
      r0[2] = 1.;  
    GLfloat proj[3];
    sub( r0, scale( assign(proj,axis), dot(axis,r0) ), r0 ); 
    normalize(r0);
    cross( axis, r0, r1 );
    for( int i=0; i<ns; i++ ) {
      GLfloat tmpC[3], tmpS[3];
      assign( tmpC, r0 );
      assign( tmpS, r1 );
      float ang = 2*M_PI/(float)ns*i;
      add( scale( tmpC, cos(ang) ), scale( tmpS, sin(ang) ), newn[i] );
      scale( assign( tmpC, newn[i] ), radius );
      add( vtx0, tmpC, npts[i] );
      n[i]  = scale( newn[i], -1. );
      pt[i] = npts[i];
    }
  }

  // do second end by translating points along axis direction
  GLfloat capn[3];
  assign( capn, axis );
  if( next ) {
    // calculate average plane of intersection
    normalize(sub( next, vtx1, capn ));
    if( dot( capn, axis)>POINT_BACK_LIMIT )  // does not point back on itself
      scale( add(axis, capn, capn), 0.5 );
  }
  float d = dot( capn, vtx1 );
  for( int i=0; i<ns; i++ ) {
    // find intersection of line through p[i] in axial 
    // direction with intersection plane
    // http://geomalgorithms.com/a05-_intersect-1.html
    GLfloat dx[3];
    GLfloat denom = dot( capn, axis );
    GLfloat num   = dot( capn, sub( pt[i], vtx1, dx ) );
    assign( dx, axis );
    add( pt[i], scale( dx, -num/denom ), npts[i+ns] );
    n[ns+i]  = n[i];
    pt[i+ns] = npts[i+ns];
  }

  // draw a cap if not joining
  if( !join ) {
    for( int i=0; i<ns; i++ ) {
      int j = (i+1)%ns;
      buff_vtx( buff,                vtx0, axis, c0, reg );
      buff_vtx( buff+1*VBO_ELEM_SZ, pt[i], axis, c0, reg );
      buff_vtx( buff+2*VBO_ELEM_SZ, pt[j], axis, c0, reg );
      buff += VBO_ELEM_SZ*3;
    }
  }

  // make cylinder elements
  for( int i=0; i<ns; i++ ) {
    int j=(i+1)%ns;
    buff_vtx( buff,               pt[j],    n[j],    c0, reg );
    buff_vtx( buff+1*VBO_ELEM_SZ, pt[i],    n[i],    c0, reg );
    buff_vtx( buff+2*VBO_ELEM_SZ, pt[ns+j], n[ns+j], c1, reg );
    buff_vtx( buff+3*VBO_ELEM_SZ, pt[i],    n[i],    c0, reg );
    buff_vtx( buff+4*VBO_ELEM_SZ, pt[ns+i], n[ns+i], c1, reg );
    buff_vtx( buff+5*VBO_ELEM_SZ, pt[ns+j], n[ns+j], c1, reg );
    buff += VBO_ELEM_SZ*6;
  }

  // draw end cap?
  if( next==NULL ){
    scale( axis, -1. );
    for( int i=0; i<ns; i++ ) {
      int j = (i+1)%ns+ns;
      buff_vtx( buff,               vtx1,     axis, c1, reg );
      buff_vtx( buff+1*VBO_ELEM_SZ, pt[i+ns], axis, c1, reg );
      buff_vtx( buff+2*VBO_ELEM_SZ, pt[j],    axis, c1, reg );
      buff += VBO_ELEM_SZ*3;
    }
    scale( axis, -1. );
  }
  int n_vtx = 3*ns*( (join?0:1) + 2 + (next==NULL?1:0) );
  return n_vtx * VBO_ELEM_SZ;
}


/**
 * @brief draw a cone/disk in x,y plane
 *
 * @param buff   VBO buffer
 * @param ndiv   number of subdivisions
 * @param rad    radius
 * @param zb     height of base
 * @param zp     height of peak
 * @param nz     normal (+/- 1)z
 * @param colour r,g,b,a
 *
 * \return number of floats buffered
 *
 * \post buff is advanced to one past end of cone/disk
 */
int
buff_cone( GLfloat* &buff, int ndiv, GLfloat rad, float zb, float zp, GLfloat nz, const GLfloat* colour ) 
{
  GLfloat x[ndiv+1], y[ndiv+1];
  GLfloat ang = 2*M_PI/float(ndiv);
  for( int i=0; i<ndiv+1; i++ ) {
    x[i] = cos( ang*i );
    y[i] = sin( ang*i );
  }

  GLfloat c[3] = { 0, 0, zp };
  GLfloat zn = copysign( rad/sqrt(rad*rad+(zb-zp)*(zb-zp)), nz );
  GLfloat xyn = sqrt(1.-zn*zn);
  GLfloat n[3];
  for( int i=0; i<ndiv; i++ )  {
    GLfloat a[3] = {0, 0, zb };
    a[0] = x[i+(nz<0)]*rad;
    a[1] = y[i+(nz<0)]*rad;
    GLfloat b[3] = {0, 0, zb };
    b[0] = x[(i+(nz>0))]*rad;
    b[1] = y[(i+(nz>0))]*rad;
    GLfloat xm = (x[i]+x[(i+1)])/2;
    GLfloat ym = (y[i]+y[(i+1)])/2;
    scale( normalize( assign( n, xm, ym, 0 ) ), xyn );
    n[2] = zn;
    scale( n, -1 );
    buff += buff_vtx( buff, c, n, colour, 0 );
    assign( n, x[i+(nz<0)]*xyn, y[i+(nz<0)]*xyn, zn ); 
    scale( n, -1 );
    buff += buff_vtx( buff, a, n, colour, 0 );
    assign( n, x[i+(nz>0)]*xyn, y[i+(nz>0)]*xyn, zn ); 
    scale( n, -1 );
    buff += buff_vtx( buff, b, n, colour, 0 );
  }
  return 3*ndiv*VBO_ELEM_SZ;
}


/**
 * @brief  fill buffer to draw an arrow centered at (0,0,0) and extending along z
 *
 * @param buffer       buffer to fill
 * @param stick        stick length
 * @param head         head length (=0 for no head)
 * @param stick_rad    stick radius
 * @param head_rad     head radius
 * @param numdiv       number of radial subdivisions
 * @param color        colour of arrow
 *
 * @post buffer filled
 * \pre  buffer must be allocated to hold arrow
 *
 * \return number of GLfloats buffered
 */
int 
buff_3Darrow( GLfloat* &buff, GLfloat stick, GLfloat head, GLfloat stick_rad, 
                    GLfloat head_rad, int numdiv, const GLfloat* colour      )
{
  GLfloat* buff0 = buff;

  GLfloat vtx0[3] = {0,0,0}, vtx1[3]={0,0,0}, vtx2[3]={0,0,0};
  vtx0[2] = -(stick+head)/2.;
  vtx1[2] =  (stick-head)/2.;
  vtx2[2] =  vtx1[2]+1;
  int nfloat = buff_cyl( buff, vtx0, vtx1, numdiv, stick_rad, colour, colour, 0, false, head?vtx2:NULL );

  if( head<0 ) return nfloat;

  // draw base of head
  buff_cone( buff, numdiv, head_rad, (stick-head)/2, (stick-head)/2., -1., colour );
  // draw head
  buff_cone( buff, numdiv, head_rad, (stick-head)/2, (stick+head)/2.,  1., colour );

  return buff - buff0;
}
