#ifndef RENDER_UTILS_H
#define RENDER_UTILS_H

#define MAX_NUM_SURFS 128

// fineness of entities drawn
#define CNNX_NUM_SIDES 8
#define SPHERE_REFINE  2


enum Sullied : unsigned int { VBO_None          = 0, 
                              VBO_View          = 1<<0, 
                              VBO_Colour        = 1<<1, 
                              VBO_Position      = 1<<2, 
                              VBO_Visible       = 1<<3, 
                              VBO_Time          = 1<<4, 
                              VBO_Hilight       = 1<<5,
                              VBO_Vector        = 1<<6,
                              VBO_Aux           = 1<<7, 
                              VBO_Lighting      = 1<<8, 
                              VBO_Clip          = 1<<9, 
                              VBO_Clip_Color    = 1<<10, 
                              VBO_Isosurf       = 1<<11, 
                              VBO_Isosurf_Color = 1<<12, 
                              VBO_Isoline       = 1<<13, 
                              VBO_Isoline_Color = 1<<14, 
                              VBO_Cnnx_Color    = 1<<15,
                              VBO_Pt_Color      = 1<<16,
                              VBO_Pt            = 1<<17,
                              VBO_VoxEle        = 1<<18,
                              VBO_All           = ~(0u) }; 

#ifdef __APPLE__
#    define GL_SILENCE_DEPRECATION
# include <OpenGL/gl3.h>
#else
#  include <FL/gl.h>
#endif
#include "Quaternion.h"
#include<vector>
#include <algorithm>    // std::rotate

#define VBO_ELEM_SZ 11
#define VBO_XYZ_OFFSET     0
#define VBO_NORM_OFFSET    3
#define VBO_COLOR_OFFSET   6
#define VBO_OPAQ_OFFSET    9
#define VBO_ENT_OFFSET    10

#define NODALBUFF_UPD_VTX 1
#define NODALBUFF_UPD_NRML 2
#define NODALBUFF_UPD_COL 4

#define XY_PLANE 1
#define XZ_PLANE 2
#define YZ_PLANE 4
#define BOGUS_VAL -666.

/**
 * @brief vertex buffer object rendering
 */
class render_data
{
    public: 
      render_data(int ps,void *c=NULL):prim_sz(ps),_context(c),_esz(ps*VBO_ELEM_SZ){
                            for(int i=0; i<MAX_NUM_SURFS;i++)set_material(i,(const float[]){BOGUS_VAL,0,0,0});}
      ~render_data();

      // light settings
      GLfloat           ambient  = 0.2f;
      GLfloat           diffuse  = 0.2f;
      GLfloat           specular = 0.7f;
      GLint             back_lit = 0;      //!< 0=cull, 1=unlit, 2=lit
      GLfloat           lineWidth = 2.;

      /**
       * @brief set properties for a surface
       *
       * @param s     surface index
       * @param props specular,diffuse,shininess,back_light
       */
      void set_material( int s, const GLfloat *props ){ 
        assert( s<MAX_NUM_SURFS );
        for( int i=0; i<4; i++ ) material[s][i]=props[i];
        _dirty = true;
      }
      
      void clear(){ nodalbuff.clear(); _n = n_opaque = 0; }
      void init_nodalbuff(bool);
      bool inited(){ return init; }
      void update_nodalbuff(){init_nodalbuff(true);}
      bool setup_render_data(void *c=NULL, bool p=false);
      virtual void mesh_render( bool wireframe=false )=0;

      int num(){return _n;}
      int numVtx(){return _n*prim_sz;}
      int prim_size(){ return prim_sz; } //!< number of vertices per element
      
      GLfloat *buffer(){ return nodalbuff.data(); }                                       
      GLfloat *transBuffer() { return nodalbuff.data() + n_opaque*VBO_ELEM_SZ;}    // portion of buffer containing translucent elements
      GLfloat *buffer(int n){ _n=n; nodalbuff.resize(n*_esz);return nodalbuff.data();}   // resize buffer for n elements
      GLfloat *buff_append(int n){_n+=n; nodalbuff.resize(_n*_esz);return nodalbuff.data()+(_n-n)*_esz;} //append n elements
      void     buff_rotate( int N ){ std::rotate( nodalbuff.begin(), 
                              nodalbuff.begin()+nodalbuff.size()-N*prim_sz*VBO_ELEM_SZ, nodalbuff.end() ); }//put last N elements at start

      void opaque( int n ){ n_opaque = prim_sz * (n<0 ? _n : n ); }          // n=number of opaque elements
      unsigned int  opaque(){ return n_opaque/prim_sz; }                     // return number opaque elements
      bool dirty(){ return _dirty; }
      void top(bool a){_on_top=a;}                                 // draw on top of all objects, unclipped
      void context( void *c ){ _context=c; }
      int num_trans(){ return _n - opaque(); }                               // number of translucent elements
      int vtx_size() { return VBO_ELEM_SZ; }                                 // number of floats/vertex

    protected:
      bool    init = false;                     //!< is structure initialized?
      int     shader_default, shader_wireframe; 
      int     prim_sz=0;                        //!< number of vertices per element
      long    _n=0;                             //!< number of elements
      long    n_opaque=-1;                      //!< number of opaque vertices to render 
      float   tfm[4][4];                        //!< transformation matrix
      vector<GLfloat> nodalbuff;               //!< data buffer
      GLuint gl_nodalbuff, gl_vertexarray;     //!< for OpenGL binding
      int    _esz;                             //!< element size
      bool   _dirty=true;                      //!< buffer needs repopulating
      void   set_clipping(int);
      void   finish_render();
      bool   _on_top=false;                    //!< draw on top?
      GLfloat material[MAX_NUM_SURFS][4];      //!< material properties for each surface
      void  *_context;                         //!< OpenGL context
      size_t _old_sz=0;                        //!< size of last buffer?
};


class RenderTris : public render_data 
{
    public:
      RenderTris(void *c=NULL):render_data(3,c){}
      void mesh_render( bool wireframe=false );
};

#define LINE_DEF_3D 4
class RenderLines : public render_data 
{
    public:
      RenderLines(void *c=NULL):render_data(1,c){sides(0);diffuse=0.7;set_material(0,defmat);}
      void mesh_render(bool=false);
      void sides( int, int=0 );
      int  sides(){ return _num_sides; }
      bool threeD() { return _num_sides>=3; }
      void threeD(bool on) {if(on==threeD())return;sides(on?LINE_DEF_3D:0);}
      /** add line of n-segments */
      void buffer_line( GLfloat* vbobuff, int n=1 );
      void adjust_n(){_n=nodalbuff.size()/VBO_ELEM_SZ;} // needed since no standard size
      RenderLines& operator += (RenderLines &rl);
    private:
      vector<int> _line_idx;            //!< vertex indices of line starts
      int         _num_sides = 0; 
      GLfloat defmat[4] = {0.7, 0.7, 80., 1.};
};


class RenderSpheres : public render_data 
{
    public:
      RenderSpheres(int ns=0,void *c=NULL):render_data(1,c){diffuse=0.7;set_material(0,defmat);}
      bool setup_render_data(void *c=NULL){ return render_data::setup_render_data(c,true); }
      void mesh_render(bool=false);
      void subdiv( int a );
      int  subdiv(){ return _nsubdiv; }
      bool threeD(){ return _nsubdiv>0; }
      void threeD( bool on ) { if(threeD()!=on) subdiv( on ? SPHERE_REFINE : 0 ); }
      void  radius( float r ){if(threeD()&&r!=_radius)_dirty=true;_radius=r; }
      float radius() { return _radius; }
      void buffer_sphere( GLfloat *, int, const GLfloat*, const GLfloat*, int );
    private:
      int              _nsubdiv=0;
      GLfloat           defmat[4] = {0.7, 0.7, 80., 1.};
      vector<GLfloat>        _vtx;  //!< icosahedron vertices
      vector<unsigned int>   _tri;  //!< icosahedron tris
      GLfloat          _radius = 1;
};

class RenderHilights : public render_data
{
    public:
        RenderHilights(void *c=NULL):render_data(1,c){diffuse=0.7;ambient=0.9;specular=0.7;set_material(0,defmat);}
        void mesh_render(bool=false);
        void pt( unsigned int, const GLfloat *, GLfloat, bool );
        void cnnx( unsigned int, const GLfloat*, const GLfloat*, GLfloat, int );
        void line( unsigned int, int p0, int p1, const int*, const GLfloat*, GLfloat, int );
        void surfele( int n, const GLfloat *, bool, int=-1 );
        void volele( int n, const GLfloat *, bool, int=-1 );
        void update_nodalbuff();
        void clear();
    private:
        GLfloat _colPt[4]       = { 1, 0.1, 0.1, 1 };
        GLfloat _colCnnx[4]     = { 1, 1, 0.5, 1 };
        GLfloat _colLine[4]     = { .1, 1, 1, 1 };
        GLfloat _colHiSurf[4]   = { .15, 0, 0.9, 1 };
        GLfloat _colAssSurf[4]  = { 1, 0.1, 1, 1 };
        GLfloat _colSurfEdge[4] = { 0, 0, 0, 1 };
        GLfloat _colHiVol[4]    = { 0.13, 0.8, 0, 1 };
        GLfloat _colAssVol[4]   = { 1.,0.55, 0.1, 1 };
        GLfloat _colVolEdge[4]  = { 0, 0, 0, 1 };
        GLfloat _pt[3];
        GLfloat _cnnx[6];
        vector<GLfloat>      _wireTri;
        unsigned int         _hiPt;
        unsigned int         _hiCnnx;
        unsigned int         _hiLine;
        unsigned int         _hiSurf;
        unsigned int         _hiVol;
        int                  _ptRad     = 1.;
        int                  _cnnxRad   = 1.;
        int                  _lineRad   = 1.;
        int                  _nPtVtx    = 1;
        int                  _nCnnxVtx  = 0;
        int                  _nLineVtx  = 0;
        bool                 _line3D    = false;
        size_t               _n_filled;
        GLfloat               defmat[4] = {1., 0.7, 80., 1.};
        void buffer_hiEles(int, const GLfloat*, bool, const GLfloat*, const GLfloat*);
};


int  generate_shader_program(const char* vertex_source, const char* fragment_source);
void align_objects();
void render_cross(float x, float y, float z, float len, float line_width);
int  buff_vtx( GLfloat *buff, const GLfloat *xyz, const GLfloat *nrml, const GLfloat *col, GLint surf );
int  buff_cyl( GLfloat*&, const GLfloat *, const GLfloat *, int ns, GLfloat, const GLfloat*, const GLfloat*, int, bool, const GLfloat* );
int  buff_sphere( GLfloat* &, const GLfloat*, const GLfloat*, int, const unsigned int*, float, const GLfloat*, int );
int  mk_icosahedron( int nsubdiv, vector<unsigned int>& tri, vector<GLfloat>& vtx );
int  buff_cone( GLfloat* &buff, int, GLfloat, GLfloat, float, GLfloat, const GLfloat* ) ;
inline int  arrow3D_ntri( int nd, bool head ){ return (head?5:4)*nd; }
int  buff_3Darrow( GLfloat* &, GLfloat, GLfloat, GLfloat, GLfloat, int, const GLfloat*);

#endif  // header guard
